# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.block_storage.v3 import group_snapshot as _group_snapshot
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import envvars
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


def _format_group_snapshot(
    snapshot: _group_snapshot.GroupSnapshot,
) -> tuple[tuple[str, ...], Iterable[Any]]:
    columns = (
        'id',
        'status',
        'name',
        'description',
        'group_id',
        'group_type_id',
    )
    column_headers = (
        'ID',
        'Status',
        'Name',
        'Description',
        'Group',
        'Group Type',
    )

    return (
        column_headers,
        utils.get_item_properties(snapshot, columns),
    )


class CreateVolumeGroupSnapshot(command.ShowOne):
    """Create a volume group snapshot.

    This command requires ``--os-volume-api-version`` 3.13 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume_group',
            metavar='<volume_group>',
            help=_('Name or ID of volume group to create a snapshot of.'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Name of the volume group snapshot.'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of a volume group snapshot.'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.14'):
            msg = _(
                "--os-volume-api-version 3.14 or greater is required to "
                "support the 'volume group snapshot create' command"
            )
            raise exceptions.CommandError(msg)

        group = volume_client.find_group(
            parsed_args.volume_group,
            ignore_missing=False,
            details=False,
        )

        snapshot = volume_client.create_group_snapshot(
            group_id=group.id,
            name=parsed_args.name,
            description=parsed_args.description,
        )

        return _format_group_snapshot(snapshot)


class DeleteVolumeGroupSnapshot(command.Command):
    """Delete a volume group snapshot.

    This command requires ``--os-volume-api-version`` 3.14 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help=_('Name or ID of volume group snapshot to delete'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.14'):
            msg = _(
                "--os-volume-api-version 3.14 or greater is required to "
                "support the 'volume group snapshot delete' command"
            )
            raise exceptions.CommandError(msg)

        group_snapshot = volume_client.find_group_snapshot(
            parsed_args.snapshot,
            ignore_missing=False,
            details=False,
        )

        volume_client.delete_group_snapshot(group_snapshot.id)


class ListVolumeGroupSnapshot(command.Lister):
    """Lists all volume group snapshot.

    This command requires ``--os-volume-api-version`` 3.14 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--all-projects',
            dest='all_projects',
            action='store_true',
            default=envvars.boolenv('ALL_PROJECTS'),
            help=_('Shows details for all projects (admin only).'),
        )
        # TODO(stephenfin): Add once we have an equivalent command for
        # 'cinder list-filters'
        # parser.add_argument(
        #     '--filter',
        #     metavar='<key=value>',
        #     action=parseractions.KeyValueAction,
        #     dest='filters',
        #     help=_(
        #         "Filter key and value pairs. Use 'foo' to "
        #         "check enabled filters from server. Use 'key~=value' for "
        #         "inexact filtering if the key supports "
        #         "(supported by --os-volume-api-version 3.33 or above)"
        #     ),
        # )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.14'):
            msg = _(
                "--os-volume-api-version 3.14 or greater is required to "
                "support the 'volume group snapshot list' command"
            )
            raise exceptions.CommandError(msg)

        groups = volume_client.group_snapshots(
            all_projects=parsed_args.all_projects,
        )

        column_headers = (
            'ID',
            'Status',
            'Name',
        )
        columns = (
            'id',
            'status',
            'name',
        )

        return (
            column_headers,
            (utils.get_item_properties(a, columns) for a in groups),
        )


class ShowVolumeGroupSnapshot(command.ShowOne):
    """Show detailed information for a volume group snapshot.

    This command requires ``--os-volume-api-version`` 3.14 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'snapshot',
            metavar='<snapshot>',
            help=_('Name or ID of volume group snapshot.'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.14'):
            msg = _(
                "--os-volume-api-version 3.14 or greater is required to "
                "support the 'volume group snapshot show' command"
            )
            raise exceptions.CommandError(msg)

        group_snapshot = volume_client.find_group_snapshot(
            parsed_args.snapshot,
            ignore_missing=False,
            details=True,
        )

        return _format_group_snapshot(group_snapshot)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from keystoneauth1 import exceptions as ks_exc
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.identity.v2_0 import role
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes


class TestRole(identity_fakes.TestIdentity):
    attr = {}
    attr['endpoints'] = [
        {
            'publicURL': identity_fakes.ENDPOINT['publicurl'],
        },
    ]
    fake_service = identity_fakes.FakeService.create_one_service(attr)
    fake_role = identity_fakes.FakeRole.create_one_role()
    fake_project = identity_fakes.FakeProject.create_one_project()
    attr = {}
    attr = {
        'tenantId': fake_project.id,
    }
    fake_user = identity_fakes.FakeUser.create_one_user(attr)

    def setUp(self):
        super().setUp()

        # Get a shortcut to the TenantManager Mock
        self.projects_mock = self.identity_client.tenants
        self.projects_mock.reset_mock()

        # Get a shortcut to the UserManager Mock
        self.users_mock = self.identity_client.users
        self.users_mock.reset_mock()

        # Get a shortcut to the RoleManager Mock
        self.roles_mock = self.identity_client.roles
        self.roles_mock.reset_mock()

        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN,
            fake_service=self.fake_service,
        )
        self.app.client_manager.auth_ref = auth_ref


class TestRoleAdd(TestRole):
    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_project

        self.users_mock.get.return_value = self.fake_user

        self.roles_mock.get.return_value = self.fake_role
        self.roles_mock.add_user_role.return_value = self.fake_role

        # Get the command object to test
        self.cmd = role.AddRole(self.app, None)

    def test_role_add(self):
        arglist = [
            '--project',
            self.fake_project.name,
            '--user',
            self.fake_user.name,
            self.fake_role.name,
        ]
        verifylist = [
            ('project', self.fake_project.name),
            ('user', self.fake_user.name),
            ('role', self.fake_role.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # RoleManager.add_user_role(user, role, tenant=None)
        self.roles_mock.add_user_role.assert_called_with(
            self.fake_user.id,
            self.fake_role.id,
            self.fake_project.id,
        )

        collist = ('id', 'name')
        self.assertEqual(collist, columns)
        datalist = (
            self.fake_role.id,
            self.fake_role.name,
        )
        self.assertEqual(datalist, data)


class TestRoleCreate(TestRole):
    fake_role_c = identity_fakes.FakeRole.create_one_role()
    columns = ('id', 'name')
    datalist = (
        fake_role_c.id,
        fake_role_c.name,
    )

    def setUp(self):
        super().setUp()

        self.roles_mock.create.return_value = self.fake_role_c

        # Get the command object to test
        self.cmd = role.CreateRole(self.app, None)

    def test_role_create_no_options(self):
        arglist = [
            self.fake_role_c.name,
        ]
        verifylist = [
            ('role_name', self.fake_role_c.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # RoleManager.create(name)
        self.roles_mock.create.assert_called_with(
            self.fake_role_c.name,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_role_create_or_show_exists(self):
        def _raise_conflict(*args, **kwargs):
            raise ks_exc.Conflict(None)

        # need to make this throw an exception...
        self.roles_mock.create.side_effect = _raise_conflict

        self.roles_mock.get.return_value = self.fake_role_c

        arglist = [
            '--or-show',
            self.fake_role_c.name,
        ]
        verifylist = [
            ('role_name', self.fake_role_c.name),
            ('or_show', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # RoleManager.get(name, description, enabled)
        self.roles_mock.get.assert_called_with(self.fake_role_c.name)

        # RoleManager.create(name)
        self.roles_mock.create.assert_called_with(
            self.fake_role_c.name,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_role_create_or_show_not_exists(self):
        arglist = [
            '--or-show',
            self.fake_role_c.name,
        ]
        verifylist = [
            ('role_name', self.fake_role_c.name),
            ('or_show', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # RoleManager.create(name)
        self.roles_mock.create.assert_called_with(
            self.fake_role_c.name,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestRoleDelete(TestRole):
    def setUp(self):
        super().setUp()

        self.roles_mock.get.return_value = self.fake_role
        self.roles_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = role.DeleteRole(self.app, None)

    def test_role_delete_no_options(self):
        arglist = [
            self.fake_role.name,
        ]
        verifylist = [
            ('roles', [self.fake_role.name]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.roles_mock.delete.assert_called_with(
            self.fake_role.id,
        )
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_delete_multi_roles_with_exception(self, find_mock):
        find_mock.side_effect = [self.fake_role, exceptions.CommandError]
        arglist = [
            self.fake_role.id,
            'unexist_role',
        ]
        verifylist = [
            ('roles', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 roles failed to delete.', str(e))

        find_mock.assert_any_call(self.roles_mock, self.fake_role.id)
        find_mock.assert_any_call(self.roles_mock, 'unexist_role')

        self.assertEqual(2, find_mock.call_count)
        self.roles_mock.delete.assert_called_once_with(self.fake_role.id)


class TestRoleList(TestRole):
    def setUp(self):
        super().setUp()

        self.roles_mock.list.return_value = [self.fake_role]

        # Get the command object to test
        self.cmd = role.ListRole(self.app, None)

    def test_role_list(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.roles_mock.list.assert_called_with()

        collist = ('ID', 'Name')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.fake_role.id,
                self.fake_role.name,
            ),
        )
        self.assertEqual(datalist, tuple(data))


class TestRoleRemove(TestRole):
    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_project

        self.users_mock.get.return_value = self.fake_user

        self.roles_mock.get.return_value = self.fake_role
        self.roles_mock.remove_user_role.return_value = None

        # Get the command object to test
        self.cmd = role.RemoveRole(self.app, None)

    def test_role_remove(self):
        arglist = [
            '--project',
            self.fake_project.name,
            '--user',
            self.fake_user.name,
            self.fake_role.name,
        ]
        verifylist = [
            ('role', self.fake_role.name),
            ('project', self.fake_project.name),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # RoleManager.remove_user_role(user, role, tenant=None)
        self.roles_mock.remove_user_role.assert_called_with(
            self.fake_user.id,
            self.fake_role.id,
            self.fake_project.id,
        )
        self.assertIsNone(result)


class TestRoleShow(TestRole):
    def setUp(self):
        super().setUp()

        self.roles_mock.get.return_value = self.fake_role

        # Get the command object to test
        self.cmd = role.ShowRole(self.app, None)

    def test_service_show(self):
        arglist = [
            self.fake_role.name,
        ]
        verifylist = [
            ('role', self.fake_role.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # RoleManager.get(role)
        self.roles_mock.get.assert_called_with(
            self.fake_role.name,
        )

        collist = ('id', 'name')
        self.assertEqual(collist, columns)
        datalist = (
            self.fake_role.id,
            self.fake_role.name,
        )
        self.assertEqual(datalist, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Floating IP Port Forwarding action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack.network.v2 import port_forwarding as _port_forwarding
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.network import common

LOG = logging.getLogger(__name__)


def validate_ports_diff(ports: list[int]) -> int:
    if len(ports) == 0:
        return 0

    ports_diff = ports[-1] - ports[0]
    if ports_diff < 0:
        msg = _(
            "The last number in port range must be"
            " greater or equal to the first"
        )
        raise exceptions.CommandError(msg)
    return ports_diff


def validate_ports_match(
    internal_ports: list[int], external_ports: list[int]
) -> None:
    internal_ports_diff = validate_ports_diff(internal_ports)
    external_ports_diff = validate_ports_diff(external_ports)

    if internal_ports_diff != 0 and internal_ports_diff != external_ports_diff:
        msg = _(
            "The relation between internal and external ports does not "
            "match the pattern 1:N and N:N"
        )
        raise exceptions.CommandError(msg)


def validate_and_assign_port_ranges(
    parsed_args: argparse.Namespace, attrs: dict[str, Any]
) -> None:
    internal_port_range = parsed_args.internal_protocol_port
    external_port_range = parsed_args.external_protocol_port
    external_ports = internal_ports = []
    if external_port_range:
        external_ports = list(map(int, str(external_port_range).split(':')))
    if internal_port_range:
        internal_ports = list(map(int, str(internal_port_range).split(':')))

    validate_ports_match(internal_ports, external_ports)

    for port in external_ports + internal_ports:
        validate_port(port)

    if internal_port_range:
        if ':' in internal_port_range:
            attrs['internal_port_range'] = internal_port_range
        else:
            attrs['internal_port'] = int(internal_port_range)

    if external_port_range:
        if ':' in external_port_range:
            attrs['external_port_range'] = external_port_range
        else:
            attrs['external_port'] = int(external_port_range)


def validate_port(port: int) -> None:
    if port <= 0 or port > 65535:
        msg = _("The port number range is <1-65535>")
        raise exceptions.CommandError(msg)


def _get_columns(
    item: _port_forwarding.PortForwarding,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class CreateFloatingIPPortForwarding(
    command.ShowOne, common.NeutronCommandWithExtraArgs
):
    _description = _("Create floating IP port forwarding")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--internal-ip-address',
            required=True,
            metavar='<internal-ip-address>',
            help=_(
                "The fixed IPv4 address of the network "
                "port associated to the floating IP port forwarding"
            ),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            required=True,
            help=_(
                "The name or ID of the network port associated "
                "to the floating IP port forwarding"
            ),
        )
        parser.add_argument(
            '--internal-protocol-port',
            metavar='<port-number>',
            required=True,
            help=_(
                "The protocol port number "
                "of the network port fixed IPv4 address "
                "associated to the floating IP port forwarding"
            ),
        )
        parser.add_argument(
            '--external-protocol-port',
            metavar='<port-number>',
            required=True,
            help=_(
                "The protocol port number of "
                "the port forwarding's floating IP address"
            ),
        )
        parser.add_argument(
            '--protocol',
            metavar='<protocol>',
            required=True,
            help=_(
                "The protocol used in the floating IP "
                "port forwarding, for instance: TCP, UDP"
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_(
                "Text to describe/contextualize the use of the "
                "port forwarding configuration"
            ),
        )
        parser.add_argument(
            'floating_ip',
            metavar='<floating-ip>',
            help=_(
                "Floating IP that the port forwarding belongs to "
                "(IP address or ID)"
            ),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        attrs: dict[str, Any] = {}
        client = self.app.client_manager.network
        floating_ip = client.find_ip(
            parsed_args.floating_ip,
            ignore_missing=False,
        )

        validate_and_assign_port_ranges(parsed_args, attrs)

        if parsed_args.port:
            port = client.find_port(parsed_args.port, ignore_missing=False)
            attrs['internal_port_id'] = port.id
        attrs['internal_ip_address'] = parsed_args.internal_ip_address
        attrs['protocol'] = parsed_args.protocol

        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description

        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        obj = client.create_floating_ip_port_forwarding(
            floating_ip.id, **attrs
        )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)


class DeleteFloatingIPPortForwarding(command.Command):
    _description = _("Delete floating IP port forwarding")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar='<floating-ip>',
            help=_(
                "Floating IP that the port forwarding belongs to "
                "(IP address or ID)"
            ),
        )
        parser.add_argument(
            'port_forwarding_id',
            nargs="+",
            metavar="<port-forwarding-id>",
            help=_("The ID of the floating IP port forwarding(s) to delete"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        floating_ip = client.find_ip(
            parsed_args.floating_ip,
            ignore_missing=False,
        )
        result = 0

        for port_forwarding_id in parsed_args.port_forwarding_id:
            try:
                client.delete_floating_ip_port_forwarding(
                    floating_ip.id,
                    port_forwarding_id,
                    ignore_missing=False,
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete floating IP port forwarding "
                        "'%(port_forwarding_id)s': %(e)s"
                    ),
                    {'port_forwarding_id': port_forwarding_id, 'e': e},
                )
        if result > 0:
            total = len(parsed_args.port_forwarding_id)
            msg = _(
                "%(result)s of %(total)s Port forwarding failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListFloatingIPPortForwarding(command.Lister):
    _description = _("List floating IP port forwarding")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar='<floating-ip>',
            help=_(
                "Floating IP that the port forwarding belongs to "
                "(IP address or ID)"
            ),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help=_(
                "List only floating IP port forwardings with the "
                "specified internal network port (name or ID)"
            ),
        )
        parser.add_argument(
            '--external-protocol-port',
            metavar='<port-number>',
            dest='external_protocol_port',
            help=_(
                "List only floating IP port forwardings with the "
                "specified external protocol port number"
            ),
        )
        parser.add_argument(
            '--protocol',
            metavar='<protocol>',
            help=_(
                "List only floating IP port forwardings with the "
                "specified protocol number"
            ),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network

        columns = (
            'id',
            'internal_port_id',
            'internal_ip_address',
            'internal_port',
            'internal_port_range',
            'external_port',
            'external_port_range',
            'protocol',
            'description',
        )
        headers = (
            'ID',
            'Internal Port ID',
            'Internal IP Address',
            'Internal Port',
            'Internal Port Range',
            'External Port',
            'External Port Range',
            'Protocol',
            'Description',
        )

        query: dict[str, Any] = {}

        if parsed_args.port:
            port = client.find_port(parsed_args.port, ignore_missing=False)
            query['internal_port_id'] = port.id
        external_port = parsed_args.external_protocol_port
        if external_port:
            if ':' in external_port:
                query['external_port_range'] = external_port
            else:
                query['external_port'] = int(
                    parsed_args.external_protocol_port
                )
        if parsed_args.protocol is not None:
            query['protocol'] = parsed_args.protocol
        if parsed_args.marker is not None:
            query['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            query['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            query['max_items'] = parsed_args.max_items

        obj = client.find_ip(
            parsed_args.floating_ip,
            ignore_missing=False,
        )

        data = client.floating_ip_port_forwardings(obj, **query)

        return (
            headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetFloatingIPPortForwarding(common.NeutronCommandWithExtraArgs):
    _description = _("Set floating IP Port Forwarding Properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar='<floating-ip>',
            help=_(
                "Floating IP that the port forwarding belongs to "
                "(IP address or ID)"
            ),
        )
        parser.add_argument(
            'port_forwarding_id',
            metavar='<port-forwarding-id>',
            help=_("The ID of the floating IP port forwarding"),
        )
        parser.add_argument(
            '--port',
            metavar='<port>',
            help=_(
                "The ID of the network port associated to "
                "the floating IP port forwarding"
            ),
        )
        parser.add_argument(
            '--internal-ip-address',
            metavar='<internal-ip-address>',
            help=_(
                "The fixed IPv4 address of the network port "
                "associated to the floating IP port forwarding"
            ),
        )
        parser.add_argument(
            '--internal-protocol-port',
            metavar='<port-number>',
            help=_(
                "The TCP/UDP/other protocol port number of the "
                "network port fixed IPv4 address associated to "
                "the floating IP port forwarding"
            ),
        )
        parser.add_argument(
            '--external-protocol-port',
            metavar='<port-number>',
            help=_(
                "The TCP/UDP/other protocol port number of the "
                "port forwarding's floating IP address"
            ),
        )
        parser.add_argument(
            '--protocol',
            metavar='<protocol>',
            choices=['tcp', 'udp'],
            help=_("The IP protocol used in the floating IP port forwarding"),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_(
                "Text to describe/contextualize the use of "
                "the port forwarding configuration"
            ),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        floating_ip = client.find_ip(
            parsed_args.floating_ip,
            ignore_missing=False,
        )

        attrs = {}
        if parsed_args.port:
            port = client.find_port(parsed_args.port, ignore_missing=False)
            attrs['internal_port_id'] = port.id

        if parsed_args.internal_ip_address:
            attrs['internal_ip_address'] = parsed_args.internal_ip_address

        validate_and_assign_port_ranges(parsed_args, attrs)

        if parsed_args.protocol:
            attrs['protocol'] = parsed_args.protocol

        if parsed_args.description is not None:
            attrs['description'] = parsed_args.description

        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )

        client.update_floating_ip_port_forwarding(
            floating_ip.id, parsed_args.port_forwarding_id, **attrs
        )


class ShowFloatingIPPortForwarding(command.ShowOne):
    _description = _("Display floating IP Port Forwarding details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar='<floating-ip>',
            help=_(
                "Floating IP that the port forwarding belongs to "
                "(IP address or ID)"
            ),
        )
        parser.add_argument(
            'port_forwarding_id',
            metavar="<port-forwarding-id>",
            help=_("The ID of the floating IP port forwarding"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        floating_ip = client.find_ip(
            parsed_args.floating_ip,
            ignore_missing=False,
        )
        obj = client.find_floating_ip_port_forwarding(
            floating_ip,
            parsed_args.port_forwarding_id,
            ignore_missing=False,
        )
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return (display_columns, data)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume v2 Type action implementations"""

import argparse
from collections.abc import MutableMapping
import functools
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from cliff import columns as cliff_columns
from openstack import utils as sdk_utils
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)


def _format_type(volume_type: Any) -> dict[str, Any]:
    ignored_columns = {'location'}
    info = volume_type.to_dict()
    data: dict[str, Any] = {}
    for key, value in info.items():
        if key in ignored_columns:
            continue
        data[key] = value

    data['properties'] = format_columns.DictColumn(data.pop('extra_specs', {}))

    return data


def _format_encryption(encryption: Any) -> dict[str, Any]:
    ignored_columns = {
        'id',
        'location',
        'name',
        'deleted',
        'created_at',
        'updated_at',
        'deleted_at',
        'volume_type_id',
    }
    info = encryption.to_dict()
    data: dict[str, Any] = {}
    for key, value in info.items():
        if key in ignored_columns:
            continue
        data[key] = value
    return data


class EncryptionInfoColumn(cliff_columns.FormattableColumn[Any]):
    """Formattable column for encryption info column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes encryption_data as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(EncryptionInfoColumn encryption_data)``.
    """

    def __init__(
        self, value: Any, encryption_data: dict[str, Any] | None = None
    ) -> None:
        super().__init__(value)
        self._encryption_data = encryption_data or {}

    def _get_encryption_info(self) -> Any:
        type_id = self._value
        return self._encryption_data.get(type_id)

    def human_readable(self) -> str:
        encryption_info = self._get_encryption_info()
        if encryption_info:
            return utils.format_dict(encryption_info)
        else:
            return '-'

    def machine_readable(self) -> Any:
        return self._get_encryption_info()


def _create_encryption_type(
    volume_client: Any, volume_type: Any, parsed_args: argparse.Namespace
) -> Any:
    if not parsed_args.encryption_provider:
        msg = _(
            "'--encryption-provider' should be specified while "
            "creating a new encryption type"
        )
        raise exceptions.CommandError(msg)

    # set the default of control location while creating
    control_location = 'front-end'
    if parsed_args.encryption_control_location:
        control_location = parsed_args.encryption_control_location

    return volume_client.create_type_encryption(
        volume_type,
        provider=parsed_args.encryption_provider,
        cipher=parsed_args.encryption_cipher,
        key_size=parsed_args.encryption_key_size,
        control_location=control_location,
    )


def _set_encryption_type(
    volume_client: Any, volume_type: Any, parsed_args: argparse.Namespace
) -> None:
    # update the existing encryption type
    kwargs = {}
    for attr in ['provider', 'cipher', 'key_size', 'control_location']:
        info = getattr(parsed_args, 'encryption_' + attr, None)
        if info is not None:
            kwargs[attr] = info
    try:
        volume_client.update_type_encryption(
            encryption=None,
            volume_type=volume_type,
            **kwargs,
        )
    except Exception as e:
        if type(e).__name__ == 'NotFound':
            # create new encryption type
            LOG.warning(
                _(
                    "No existing encryption type found, creating "
                    "new encryption type for this volume type ..."
                )
            )
            _create_encryption_type(volume_client, volume_type, parsed_args)


class CreateVolumeType(command.ShowOne):
    _description = _("Create new volume type")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "name",
            metavar="<name>",
            help=_("Volume type name"),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Volume type description"),
        )
        public_group = parser.add_mutually_exclusive_group()
        public_group.add_argument(
            "--public",
            action="store_true",
            dest="is_public",
            default=None,
            help=_("Volume type is accessible to the public"),
        )
        public_group.add_argument(
            "--private",
            action="store_false",
            dest="is_public",
            default=None,
            help=_("Volume type is not accessible to the public"),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            help=_(
                'Set a property on this volume type '
                '(repeat option to set multiple properties)'
            ),
        )
        parser.add_argument(
            '--multiattach',
            action='store_true',
            default=False,
            help=_(
                "Enable multi-attach for this volume type "
                "(this is an alias for '--property multiattach=<is> True') "
                "(requires driver support)"
            ),
        )
        parser.add_argument(
            '--cacheable',
            action='store_true',
            default=False,
            help=_(
                "Enable caching for this volume type "
                "(this is an alias for '--property cacheable=<is> True') "
                "(requires driver support)"
            ),
        )
        parser.add_argument(
            '--replicated',
            action='store_true',
            default=False,
            help=_(
                "Enabled replication for this volume type "
                "(this is an alias for "
                "'--property replication_enabled=<is> True') "
                "(requires driver support)"
            ),
        )
        parser.add_argument(
            '--availability-zone',
            action='append',
            dest='availability_zones',
            help=_(
                "Set an availability zone for this volume type "
                "(this is an alias for "
                "'--property RESKEY:availability_zones:<az>') "
                "(repeat option to set multiple availability zones)"
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "Allow <project> to access private type (name or ID) "
                "(must be used with --private option)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        # TODO(Huanxuan Ao): Add choices for each "--encryption-*" option.
        parser.add_argument(
            '--encryption-provider',
            metavar='<provider>',
            help=_(
                'Set the encryption provider format for '
                'this volume type (e.g "luks" or "plain") (admin only) '
                '(this option is required when setting encryption type '
                'of a volume; consider using other encryption options '
                'such as: "--encryption-cipher", "--encryption-key-size" '
                'and "--encryption-control-location")'
            ),
        )
        parser.add_argument(
            '--encryption-cipher',
            metavar='<cipher>',
            help=_(
                'Set the encryption algorithm or mode for this '
                'volume type (e.g "aes-xts-plain64") (admin only)'
            ),
        )
        parser.add_argument(
            '--encryption-key-size',
            metavar='<key-size>',
            type=int,
            help=_(
                'Set the size of the encryption key of this '
                'volume type (e.g "128" or "256") (admin only)'
            ),
        )
        parser.add_argument(
            '--encryption-control-location',
            metavar='<control-location>',
            choices=['front-end', 'back-end'],
            help=_(
                'Set the notional service where the encryption is '
                'performed ("front-end" or "back-end") (admin only) '
                '(The default value for this option is "front-end" '
                'when setting encryption type of a volume. Consider '
                'using other encryption options such as: '
                '"--encryption-cipher", "--encryption-key-size" and '
                '"--encryption-provider")'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        # we don't narrow types here since we want to handle both v2.0 and v3
        identity_client = self.app.client_manager.sdk_connection.identity

        if parsed_args.project and parsed_args.is_public is not False:
            msg = _("--project is only allowed with --private")
            raise exceptions.CommandError(msg)

        kwargs = {}

        if parsed_args.is_public is not None:
            kwargs['is_public'] = parsed_args.is_public

        volume_type = volume_client.create_type(
            name=parsed_args.name,
            description=parsed_args.description,
            **kwargs,
        )

        info = _format_type(volume_type)

        if parsed_args.project:
            try:
                project_id = identity_common.find_project_id_sdk(
                    identity_client,
                    parsed_args.project,
                    parsed_args.project_domain,
                )
                volume_client.add_type_access(volume_type.id, project_id)
            except Exception as e:
                msg = _(
                    "Failed to add project %(project)s access to type: %(e)s"
                )
                LOG.error(msg, {'project': parsed_args.project, 'e': e})

        properties = {}
        if parsed_args.properties:
            properties.update(parsed_args.properties)
        if parsed_args.multiattach:
            properties['multiattach'] = '<is> True'
        if parsed_args.cacheable:
            properties['cacheable'] = '<is> True'
        if parsed_args.replicated:
            properties['replication_enabled'] = '<is> True'
        if parsed_args.availability_zones:
            properties['RESKEY:availability_zones'] = ','.join(
                parsed_args.availability_zones
            )
        if properties:
            result = volume_client.update_type_extra_specs(
                volume_type.id, **properties
            )
            info['properties'] = format_columns.DictColumn(result.extra_specs)

        if (
            parsed_args.encryption_provider
            or parsed_args.encryption_cipher
            or parsed_args.encryption_key_size
            or parsed_args.encryption_control_location
        ):
            try:
                # create new encryption
                encryption = _create_encryption_type(
                    volume_client, volume_type, parsed_args
                )
            except Exception as e:
                LOG.error(
                    _(
                        "Failed to set encryption information for this "
                        "volume type: %s"
                    ),
                    e,
                )
            # add encryption info in result
            encryption_info = _format_encryption(encryption)
            info['encryption'] = format_columns.DictColumn(encryption_info)

        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data


class DeleteVolumeType(command.Command):
    _description = _("Delete volume type(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "volume_types",
            metavar="<volume-type>",
            nargs="+",
            help=_("Volume type(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        result = 0

        for volume_type in parsed_args.volume_types:
            try:
                vol_type = volume_client.find_type(
                    volume_type, ignore_missing=False
                )
                volume_client.delete_type(vol_type)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete volume type with "
                        "name or ID '%(volume_type)s': %(e)s"
                    ),
                    {'volume_type': volume_type, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.volume_types)
            msg = _(
                "%(result)s of %(total)s volume types failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListVolumeType(command.Lister):
    _description = _("List volume types")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        public_group = parser.add_mutually_exclusive_group()
        public_group.add_argument(
            "--default",
            action='store_true',
            default=False,
            help=_('List the default volume type'),
        )
        public_group.add_argument(
            "--public",
            action="store_true",
            dest="is_public",
            default='none',  # cinder expects the string "none"
            help=_("List only public types"),
        )
        public_group.add_argument(
            "--private",
            action="store_false",
            dest="is_public",
            default='none',  # cinder expects the string "none"
            help=_("List only private types (admin only)"),
        )
        parser.add_argument(
            "--encryption-type",
            action="store_true",
            help=_(
                "Display encryption information for each volume type "
                "(admin only)"
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )

        if parsed_args.long:
            columns = [
                'ID',
                'Name',
                'Is Public',
                'Description',
            ]
            column_headers = [
                'ID',
                'Name',
                'Is Public',
                'Description',
            ]
        else:
            columns = ['ID', 'Name', 'Is Public']
            column_headers = ['ID', 'Name', 'Is Public']

        if parsed_args.default:
            data = [volume_client.get_type('default')]
        else:
            data = list(
                volume_client.types(
                    is_public=parsed_args.is_public,
                )
            )

        formatters: MutableMapping[str, Any] = {
            'Extra Specs': format_columns.DictColumn
        }

        if parsed_args.encryption_type:
            encryption = {}
            for d in data:
                e = volume_client.get_type_encryption(d.id)
                if not e.volume_type_id:
                    continue

                encryption_info = _format_encryption(e)
                encryption[e.volume_type_id] = encryption_info

            # We need to get volume type ID, then show encryption
            # information according to the ID, so use "id" to keep
            # difference to the real "ID" column.
            columns += ['id']
            column_headers += ['Encryption']

            _EncryptionInfoColumn = functools.partial(
                EncryptionInfoColumn, encryption_data=encryption
            )
            formatters['id'] = _EncryptionInfoColumn

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters=formatters,
                )
                for s in data
            ),
        )


class SetVolumeType(command.Command):
    _description = _("Set volume type properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume_type',
            metavar='<volume-type>',
            help=_('Volume type to modify (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Set volume type name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Set volume type description'),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            help=_(
                'Set a property on this volume type '
                '(repeat option to set multiple properties)'
            ),
        )
        parser.add_argument(
            '--multiattach',
            action='store_true',
            default=False,
            help=_(
                "Enable multi-attach for this volume type "
                "(this is an alias for '--property multiattach=<is> True') "
                "(requires driver support)"
            ),
        )
        parser.add_argument(
            '--cacheable',
            action='store_true',
            default=False,
            help=_(
                "Enable caching for this volume type "
                "(this is an alias for '--property cacheable=<is> True') "
                "(requires driver support)"
            ),
        )
        parser.add_argument(
            '--replicated',
            action='store_true',
            default=False,
            help=_(
                "Enabled replication for this volume type "
                "(this is an alias for "
                "'--property replication_enabled=<is> True') "
                "(requires driver support)"
            ),
        )
        parser.add_argument(
            '--availability-zone',
            action='append',
            dest='availability_zones',
            help=_(
                "Set an availability zone for this volume type "
                "(this is an alias for "
                "'--property RESKEY:availability_zones:<az>') "
                "(repeat option to set multiple availability zones)"
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Set volume type access to project (name or ID) (admin only)'
            ),
        )
        public_group = parser.add_mutually_exclusive_group()
        public_group.add_argument(
            '--public',
            action='store_true',
            dest='is_public',
            default=None,
            help=_('Volume type is accessible to the public'),
        )
        public_group.add_argument(
            '--private',
            action='store_false',
            dest='is_public',
            default=None,
            help=_("Volume type is not accessible to the public"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        # TODO(Huanxuan Ao): Add choices for each "--encryption-*" option.
        parser.add_argument(
            '--encryption-provider',
            metavar='<provider>',
            help=_(
                'Set the encryption provider format for '
                'this volume type (e.g "luks" or "plain") (admin only) '
                '(This option is required when setting encryption type '
                'of a volume for the first time. Consider using other '
                'encryption options such as: "--encryption-cipher", '
                '"--encryption-key-size" and '
                '"--encryption-control-location")'
            ),
        )
        parser.add_argument(
            '--encryption-cipher',
            metavar='<cipher>',
            help=_(
                'Set the encryption algorithm or mode for this '
                'volume type (e.g "aes-xts-plain64") (admin only)'
            ),
        )
        parser.add_argument(
            '--encryption-key-size',
            metavar='<key-size>',
            type=int,
            help=_(
                'Set the size of the encryption key of this '
                'volume type (e.g "128" or "256") (admin only)'
            ),
        )
        parser.add_argument(
            '--encryption-control-location',
            metavar='<control-location>',
            choices=['front-end', 'back-end'],
            help=_(
                'Set the notional service where the encryption is '
                'performed ("front-end" or "back-end") (admin only) '
                '(The default value for this option is "front-end" '
                'when setting encryption type of a volume for the '
                'first time. Consider using other encryption options '
                'such as: "--encryption-cipher", "--encryption-key-size" '
                'and "--encryption-provider")'
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        # we don't narrow types here since we want to handle both v2.0 and v3
        identity_client = self.app.client_manager.sdk_connection.identity

        volume_type = volume_client.find_type(
            parsed_args.volume_type, ignore_missing=False
        )

        result = 0
        kwargs = {}

        if parsed_args.name:
            kwargs['name'] = parsed_args.name

        if parsed_args.description:
            kwargs['description'] = parsed_args.description

        if parsed_args.is_public is not None:
            kwargs['is_public'] = parsed_args.is_public

        if kwargs:
            try:
                volume_client.update_type(volume_type.id, **kwargs)
            except Exception as e:
                LOG.error(
                    _("Failed to update volume type name or description: %s"),
                    e,
                )
                result += 1

        properties: dict[str, str] = {}
        if parsed_args.properties:
            properties.update(parsed_args.properties)
        if parsed_args.multiattach:
            properties['multiattach'] = '<is> True'
        if parsed_args.cacheable:
            properties['cacheable'] = '<is> True'
        if parsed_args.replicated:
            properties['replication_enabled'] = '<is> True'
        if parsed_args.availability_zones:
            properties['RESKEY:availability_zones'] = ','.join(
                parsed_args.availability_zones
            )
        if properties:
            try:
                volume_client.update_type_extra_specs(
                    volume_type.id, **properties
                )
            except Exception as e:
                LOG.error(_("Failed to set volume type properties: %s"), e)
                result += 1

        if parsed_args.project:
            try:
                project_id = identity_common.find_project_id_sdk(
                    identity_client,
                    parsed_args.project,
                    parsed_args.project_domain,
                )
                volume_client.add_type_access(volume_type.id, project_id)
            except Exception as e:
                LOG.error(
                    _("Failed to set volume type access to project: %s"), e
                )
                result += 1

        if (
            parsed_args.encryption_provider
            or parsed_args.encryption_cipher
            or parsed_args.encryption_key_size
            or parsed_args.encryption_control_location
        ):
            try:
                _set_encryption_type(volume_client, volume_type, parsed_args)
            except Exception as e:
                LOG.error(
                    _(
                        "Failed to set encryption information for this "
                        "volume type: %s"
                    ),
                    e,
                )
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("Command Failed: One or more of the operations failed")
            )


class ShowVolumeType(command.ShowOne):
    _description = _("Display volume type details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "volume_type",
            metavar="<volume-type>",
            help=_("Volume type to display (name or ID)"),
        )
        parser.add_argument(
            "--encryption-type",
            action="store_true",
            help=_(
                "Display encryption information of this volume type "
                "(admin only)"
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )

        volume_type = volume_client.find_type(
            parsed_args.volume_type, ignore_missing=False
        )

        info = _format_type(volume_type)

        access_project_ids = None
        if not volume_type.is_public:
            try:
                volume_type_access = volume_client.get_type_access(
                    volume_type.id
                )
                project_ids = [
                    utils.get_field(item, 'project_id')
                    for item in volume_type_access
                ]
                # TODO(Rui Chen): This format list case can be removed after
                # patch https://review.opendev.org/#/c/330223/ merged.
                access_project_ids = format_columns.ListColumn(project_ids)
            except Exception as e:
                msg = _(
                    'Failed to get access project list for volume type '
                    '%(type)s: %(e)s'
                )
                LOG.error(msg, {'type': volume_type.id, 'e': e})
        info['access_project_ids'] = access_project_ids

        if parsed_args.encryption_type:
            # show encryption type information for this volume type
            try:
                encryption = volume_client.get_type_encryption(volume_type.id)
                encryption_info = _format_encryption(encryption)
                info['encryption'] = format_columns.DictColumn(encryption_info)
            except Exception as e:
                LOG.error(
                    _(
                        "Failed to display the encryption information "
                        "of this volume type: %s"
                    ),
                    e,
                )
        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data


class UnsetVolumeType(command.Command):
    _description = _("Unset volume type properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume_type',
            metavar='<volume-type>',
            help=_('Volume type to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            dest='properties',
            help=_(
                'Remove a property from this volume type '
                '(repeat option to remove multiple properties)'
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Removes volume type access to project (name or ID) '
                '(admin only)'
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            "--encryption-type",
            action="store_true",
            help=_(
                "Remove the encryption type for this volume type (admin only)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '2'
        )
        # we don't narrow types here since we want to handle both v2.0 and v3
        identity_client = self.app.client_manager.sdk_connection.identity

        volume_type = volume_client.find_type(
            parsed_args.volume_type, ignore_missing=False
        )

        result = 0
        if parsed_args.properties:
            try:
                volume_client.delete_type_extra_specs(
                    volume_type.id, parsed_args.properties
                )
            except Exception as e:
                LOG.error(_("Failed to unset volume type properties: %s"), e)
                result += 1

        if parsed_args.project:
            try:
                project_id = identity_common.find_project_id_sdk(
                    identity_client,
                    parsed_args.project,
                    parsed_args.project_domain,
                )
                volume_client.remove_type_access(volume_type.id, project_id)
            except Exception as e:
                LOG.error(
                    _("Failed to remove volume type access from project: %s"),
                    e,
                )
                result += 1
        if parsed_args.encryption_type:
            try:
                volume_client.delete_type_encryption(None, volume_type.id)
            except Exception as e:
                LOG.error(
                    _(
                        "Failed to remove the encryption type for this "
                        "volume type: %s"
                    ),
                    e,
                )
                result += 1

        if result > 0:
            raise exceptions.CommandError(
                _("Command Failed: One or more of the operations failed")
            )



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from keystoneclient import exceptions as identity_exc
from osc_lib import exceptions

from openstackclient.identity.v2_0 import service
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes


class TestService(identity_fakes.TestIdentity):
    fake_service = identity_fakes.FakeService.create_one_service()

    def setUp(self):
        super().setUp()

        # Get a shortcut to the ServiceManager Mock
        self.services_mock = self.identity_client.services
        self.services_mock.reset_mock()


class TestServiceCreate(TestService):
    fake_service_c = identity_fakes.FakeService.create_one_service()
    columns = (
        'description',
        'id',
        'name',
        'type',
    )
    datalist = (
        fake_service_c.description,
        fake_service_c.id,
        fake_service_c.name,
        fake_service_c.type,
    )

    def setUp(self):
        super().setUp()

        self.services_mock.create.return_value = self.fake_service_c

        # Get the command object to test
        self.cmd = service.CreateService(self.app, None)

    def test_service_create(self):
        arglist = [
            self.fake_service_c.type,
        ]
        verifylist = [
            ('type', self.fake_service_c.type),
            ('name', None),
            ('description', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # ServiceManager.create(name, service_type, description)
        self.services_mock.create.assert_called_with(
            None,
            self.fake_service_c.type,
            None,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_service_create_with_name_option(self):
        arglist = [
            '--name',
            self.fake_service_c.name,
            self.fake_service_c.type,
        ]
        verifylist = [
            ('type', self.fake_service_c.type),
            ('name', self.fake_service_c.name),
            ('description', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # ServiceManager.create(name, service_type, description)
        self.services_mock.create.assert_called_with(
            self.fake_service_c.name,
            self.fake_service_c.type,
            None,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_service_create_description(self):
        arglist = [
            '--name',
            self.fake_service_c.name,
            '--description',
            self.fake_service_c.description,
            self.fake_service_c.type,
        ]
        verifylist = [
            ('type', self.fake_service_c.type),
            ('name', self.fake_service_c.name),
            ('description', self.fake_service_c.description),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # ServiceManager.create(name, service_type, description)
        self.services_mock.create.assert_called_with(
            self.fake_service_c.name,
            self.fake_service_c.type,
            self.fake_service_c.description,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestServiceDelete(TestService):
    def setUp(self):
        super().setUp()

        self.services_mock.get.side_effect = identity_exc.NotFound(None)
        self.services_mock.find.return_value = self.fake_service
        self.services_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = service.DeleteService(self.app, None)

    def test_service_delete_no_options(self):
        arglist = [
            self.fake_service.name,
        ]
        verifylist = [
            ('services', [self.fake_service.name]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.services_mock.delete.assert_called_with(
            self.fake_service.id,
        )
        self.assertIsNone(result)


class TestServiceList(TestService):
    def setUp(self):
        super().setUp()

        self.services_mock.list.return_value = [self.fake_service]

        # Get the command object to test
        self.cmd = service.ListService(self.app, None)

    def test_service_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.services_mock.list.assert_called_with()

        collist = ('ID', 'Name', 'Type')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.fake_service.id,
                self.fake_service.name,
                self.fake_service.type,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_service_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.services_mock.list.assert_called_with()

        collist = ('ID', 'Name', 'Type', 'Description')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.fake_service.id,
                self.fake_service.name,
                self.fake_service.type,
                self.fake_service.description,
            ),
        )
        self.assertEqual(datalist, tuple(data))


class TestServiceShow(TestService):
    fake_service_s = identity_fakes.FakeService.create_one_service()

    def setUp(self):
        super().setUp()

        self.services_mock.get.side_effect = identity_exc.NotFound(None)
        self.services_mock.find.return_value = self.fake_service_s

        # Get the command object to test
        self.cmd = service.ShowService(self.app, None)

    def test_service_show(self):
        arglist = [
            self.fake_service_s.name,
        ]
        verifylist = [
            ('service', self.fake_service_s.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # ServiceManager.find(id)
        self.services_mock.find.assert_called_with(
            name=self.fake_service_s.name,
        )

        collist = ('description', 'id', 'name', 'type')
        self.assertEqual(collist, columns)
        datalist = (
            self.fake_service_s.description,
            self.fake_service_s.id,
            self.fake_service_s.name,
            self.fake_service_s.type,
        )
        self.assertEqual(datalist, data)

    def test_service_show_nounique(self):
        self.services_mock.find.side_effect = identity_exc.NoUniqueMatch(None)
        arglist = [
            'nounique_service',
        ]
        verifylist = [
            ('service', 'nounique_service'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                "Multiple service matches found for 'nounique_service',"
                " use an ID to be more specific.",
                str(e),
            )
